import ctypes
import re
from enum import IntEnum
from typing import Annotated

import pytest

from pymhf.extensions.ctypes import c_enum32
from pymhf.utils.partial_struct import Field, partial_struct


def test_simple_structure():
    @partial_struct
    class Test(ctypes.Structure):
        a: Annotated[ctypes.c_uint32, 0x0]
        b: Annotated[ctypes.c_uint32, 0x10]

    assert Test._fields_ == [
        ("a", ctypes.c_uint32),
        ("_padding_4", ctypes.c_ubyte * 0xC),
        ("b", ctypes.c_uint32),
    ]

    data = bytearray(b"\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00\x04\x00\x00\x00\x05\x00\x00\x00")
    t = Test.from_buffer(data)
    assert t.a == 1
    assert t.b == 5
    assert bytes(t) == bytes(data)

    # Also test modifying a value.
    t.a = 42
    assert bytes(t) == b"\x2a\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00\x04\x00\x00\x00\x05\x00\x00\x00"


def test_simple_structure_with_enum():
    class Alphabet(IntEnum):
        A = 0
        B = 1
        C = 2
        D = 3
        E = 4

    @partial_struct
    class Test(ctypes.Structure):
        a: Annotated[c_enum32[Alphabet], 0x0]
        b: Annotated[ctypes.c_uint32, 0x10]

    assert Test._fields_ == [
        ("a", c_enum32[Alphabet]),
        ("_padding_4", ctypes.c_ubyte * 0xC),
        ("b", ctypes.c_uint32),
    ]

    data = bytearray(b"\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00\x04\x00\x00\x00\x05\x00\x00\x00")
    t = Test.from_buffer(data)
    assert t.a == 1
    assert t.a == Alphabet.B
    assert t.b == 5
    assert bytes(t) == bytes(data)

    # Test modifying the enum value
    t.a = Alphabet.D
    assert bytes(t) == b"\x03\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00\x04\x00\x00\x00\x05\x00\x00\x00"


def test_simple_structure_with_total_size():
    # Test case for the partial struct having a _total_size_ attribute.
    @partial_struct
    class Test(ctypes.Structure):
        _total_size_ = 0x18
        a: Annotated[int, Field(ctypes.c_uint32)]
        b: Annotated[int, Field(ctypes.c_uint32, 0x10)]

    assert Test._fields_ == [
        ("a", ctypes.c_uint32),
        ("_padding_4", ctypes.c_ubyte * 0xC),
        ("b", ctypes.c_uint32),
        ("_padding_14", ctypes.c_ubyte * 0x4),
    ]

    data = bytearray(
        b"\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00\x04\x00\x00\x00\x05\x00\x00\x00\x00\x00\x00\x00"
    )
    t = Test.from_buffer(data)
    assert t.a == 1
    assert t.b == 5
    assert bytes(t) == bytes(data)


def test_nested_structs():
    # Test for one of the fields being a struct nested within the currently defined one.
    @partial_struct
    class Test(ctypes.Structure):
        class Sub(ctypes.Structure):
            _fields_ = [
                ("sub_a", ctypes.c_uint16),
                ("sub_b", ctypes.c_uint16),
            ]

        _total_size_ = 24
        a: ctypes.c_uint32
        b_sub: Sub
        c: Annotated[ctypes.c_uint32, 0x8]
        d: Annotated[int, Field(ctypes.c_uint32, 0x10)]

    assert Test._fields_ == [
        ("a", ctypes.c_uint32),
        ("b_sub", Test.Sub),
        ("c", ctypes.c_uint32),
        ("_padding_C", ctypes.c_ubyte * 0x4),
        ("d", ctypes.c_uint32),
        ("_padding_14", ctypes.c_ubyte * 0x4),
    ]

    data = bytearray(
        b"\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00\x04\x00\x00\x00\x05\x00\x00\x00\x00\x00\x00\x00"
    )
    t = Test.from_buffer(data)
    assert t.a == 1
    assert t.b_sub.sub_a == 2
    assert t.b_sub.sub_b == 0
    assert t.c == 3
    assert t.d == 5

    assert bytes(t) == bytes(data)


def test_annotated_struct():
    # Test the case of the type being an annotation.

    class Sub(ctypes.Structure):
        _fields_ = [
            ("sub_a", ctypes.c_uint16),
            ("sub_b", ctypes.c_uint16),
        ]

    @partial_struct
    class Test(ctypes.Structure):
        _total_size_ = 24
        a: ctypes.c_uint32
        b_sub: "Sub"
        c: Annotated["Sub", 0x8]
        d: Annotated[int, Field(ctypes.c_uint32, 0x10)]

    assert Test._fields_ == [
        ("a", ctypes.c_uint32),
        ("b_sub", Sub),
        ("c", Sub),
        ("_padding_C", ctypes.c_ubyte * 0x4),
        ("d", ctypes.c_uint32),
        ("_padding_14", ctypes.c_ubyte * 0x4),
    ]

    data = bytearray(
        b"\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00\x04\x00\x00\x00\x05\x00\x00\x00\x00\x00\x00\x00"
    )
    t = Test.from_buffer(data)
    assert t.a == 1
    assert t.b_sub.sub_a == 2
    assert t.b_sub.sub_b == 0
    assert t.c.sub_a == 3
    assert t.c.sub_b == 0
    assert t.d == 5

    assert bytes(t) == bytes(data)


def test_structure_with_pointer():
    # Test a struct which has a pointer in it.
    # Also implicitly test the case of putting an "invalid" offset.
    # In this case we put the pointer at 0xC, but because we're 64 bit it has to be aligned to 0x8 byte
    # boundary.
    @partial_struct
    class Test(ctypes.Structure):
        a: Annotated[ctypes.c_uint32, 0x0]
        a_bool: Annotated[ctypes.c_bool, 0xA]
        b: Annotated[ctypes._Pointer[ctypes.c_uint32], 0xB]
        c: Annotated[ctypes.c_uint32, 0x18]

    assert Test._fields_ == [
        ("a", ctypes.c_uint32),
        ("_padding_4", ctypes.c_ubyte * 0x6),
        ("a_bool", ctypes.c_bool),
        ("b", ctypes.POINTER(ctypes.c_uint32)),
        ("c", ctypes.c_uint32),
    ]

    assert ctypes.sizeof(Test) == 0x20  # 4 extra bytes at the end since the whole struct will be 0x8 aligned.

    # Check all of our offsets are also correct in the type.
    assert Test.a.offset == 0
    assert Test.a_bool.offset == 0xA
    assert Test.b.offset == 0x10
    assert Test.c.offset == 0x18

    data = bytearray(
        b"\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x07\x00\x00\x00\x00\x00\x00\x00"
    )
    t = Test.from_buffer(data)
    assert t.a == 1
    with pytest.raises(ValueError, match="NULL pointer access"):
        t.b.contents
    assert t.c == 7
    assert bytes(t) == bytes(data)


def test_structure_with_annotated_pointer():
    # Test a struct which has an annotated pointer in it.
    class Sub(ctypes.Structure):
        _fields_ = [
            ("sub_a", ctypes.c_uint16),
            ("sub_b", ctypes.c_uint16),
        ]

    @partial_struct
    class Test(ctypes.Structure):
        a: ctypes._Pointer[Sub]
        b: "ctypes._Pointer[Sub]"

    assert Test._fields_ == [("a", ctypes.POINTER(Sub)), ("b", ctypes.POINTER(Sub))]

    assert ctypes.sizeof(Test) == 0x10

    # Check all of our offsets are also correct in the type.
    assert Test.a.offset == 0
    assert Test.b.offset == 0x8

    data = bytearray(b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00")
    t = Test.from_buffer(data)
    with pytest.raises(ValueError, match="NULL pointer access"):
        t.a.contents
    with pytest.raises(ValueError, match="NULL pointer access"):
        t.b.contents
    assert bytes(t) == bytes(data)


def test_annotated_arrays():
    # Test the case of an array of some other type.

    class Sub(ctypes.Structure):
        _fields_ = [
            ("sub_a", ctypes.c_uint16),
            ("sub_b", ctypes.c_uint16),
        ]

    @partial_struct
    class Test(ctypes.Structure):
        _total_size_ = 24
        a: ctypes.c_uint32
        b: "Sub * 2"
        c: Annotated[list[Sub], Field(Sub * 3)]

    assert Test._fields_ == [
        ("a", ctypes.c_uint32),
        ("b", Sub * 2),
        ("c", Sub * 3),
    ]

    data = bytearray(
        b"\x01\x00\x00\x00\x02\x00\x00\x00\x03\x00\x00\x00\x04\x00\x00\x00\x05\x00\x00\x00\x06\x00\x00\x00"
    )
    t = Test.from_buffer(data)
    assert t.a == 1
    assert t.b[0].sub_a == 2
    assert t.b[0].sub_b == 0
    assert t.b[1].sub_a == 3
    assert t.b[1].sub_b == 0
    assert len(t.c) == 3
    assert t.c[0].sub_a == 4
    assert t.c[0].sub_b == 0

    assert bytes(t) == bytes(data)


def test_invalid_cases():
    # Invalid type type
    with pytest.raises(ValueError, match=re.escape("The field 'a' has an invalid type: <class 'int'>")):

        @partial_struct
        class Test(ctypes.Structure):
            a: int

    with pytest.raises(ValueError, match=re.escape("The field 'a' has an invalid type: <class 'int'>")):

        @partial_struct
        class Test(ctypes.Structure):
            a: Annotated[int, int]

    # Invalid annotation
    with pytest.raises(
        ValueError,
        match=re.escape(
            "The field 'a' has an invalid annotation: typing.Annotated[int, <class 'str'>, <class 'float'>]"
        ),
    ):

        @partial_struct
        class Test(ctypes.Structure):
            a: Annotated[int, str, float]

    # Invalid offset
    with pytest.raises(ValueError, match=re.escape("The field 'a' has an invalid offset: 'hi'")):

        @partial_struct
        class Test(ctypes.Structure):
            a: Annotated[ctypes.c_int32, "hi"]


def test_invalid_c_enum32_cases():
    with pytest.raises(TypeError):
        c_enum32[22]



import ctypes
import inspect
from _ctypes import _Pointer
from functools import lru_cache
from typing import Any, Callable, NamedTuple, Optional, Union, get_args

from typing_extensions import get_type_hints

from pymhf.core._types import FUNCDEF

CTYPES = Union[ctypes._SimpleCData, ctypes.Structure, ctypes._Pointer, _Pointer]


class ArgData(NamedTuple):
    name: str
    arg_type: CTYPES


class FuncDef:
    def __init__(self, restype: Any, argtypes: list[ArgData], defaults: Optional[dict] = None):
        self.restype = restype
        self._arg_names = [x.name for x in argtypes]
        self._arg_types = [x.arg_type for x in argtypes]
        self.defaults = defaults or dict()

    @property
    def arg_types(self) -> list[CTYPES]:
        return self._arg_types

    @property
    def arg_names(self) -> list[str]:
        return self._arg_names

    def to_FUNCDEF(self) -> FUNCDEF:
        return FUNCDEF(self.restype, self.arg_types)

    def flatten(self, *args, **kwargs):
        """Take the provided signature, args and kwargs and convert to a single list of args."""
        out_args = {}
        missing = []
        # First, apply the arg values
        for i, arg in enumerate(args):
            out_args[self.arg_names[i]] = arg
        # Then, if we have any kwargs, combine with the defaults and then apply to the args.
        if len(out_args) != len(self.arg_names):
            _defaults = dict(self.defaults)
            _defaults.update(kwargs)
            # To retain order, loop over the remaining indexes and pick out the corresponding arg.
            for i in range(len(out_args), len(self.arg_names)):
                key = self.arg_names[i]
                if key in _defaults:
                    out_args[key] = _defaults[key]
                else:
                    missing.append(key)
        if missing:
            raise ValueError(f"Missing argument(s): {missing}")
        return list(out_args.values())


@lru_cache(maxsize=1024)
def _get_funcdef(func: Callable) -> FuncDef:
    """Get the funcdef for the provided function.
    This is wrapped in an lru_cache so that if multiple detours use the same function, it will only be
    analysed once."""
    func_params = inspect.signature(func).parameters
    func_type_hints = get_type_hints(func)
    _restype = func_type_hints.pop("return", type(None))
    if _restype is type(None):
        restype = None
    elif issubclass(_restype, ctypes._SimpleCData):
        restype = _restype
    else:
        raise TypeError("Return type must be a subclass of a ctypes.Structure or a simple ctypes type.")
    argtypes = []
    defaults = {}
    _missing = []
    for name, param in func_params.items():
        if name != "self":
            if name in func_type_hints:
                argtype = func_type_hints[name]
                if issubclass(argtype, get_args(CTYPES)):
                    default_val = param.default
                    if default_val != inspect.Signature.empty:
                        defaults[name] = default_val
                    argtypes.append(ArgData(name, argtype))
                else:
                    raise TypeError(f"Invalid type {argtype!r} for argument {name!r}")
            else:
                _missing.append(name)
    if _missing:
        raise TypeError(f"The argument(s) {_missing} for {func.__name__!r} do not have type hints provided.")
    return FuncDef(restype, argtypes, defaults)



import re
from typing import Optional

import tomlkit

REGEX = r"(?m)^# /// (?P<type>[a-zA-Z0-9-]+)$\s(?P<content>(^#(| .*)$\s)+)^# ///$"


def read_inline_metadata(script: str) -> Optional[tomlkit.TOMLDocument]:
    """Read a file and extract the toml info contained in the script if there is one.
    This is taken directly from the reference implementation in https://peps.python.org/pep-0723/
    and modified to use `tomlkit` to ensure compatibility for python 3.9+
    """
    name = "script"
    matches = list(filter(lambda m: m.group("type") == name, re.finditer(REGEX, script)))
    if len(matches) > 1:
        raise ValueError(f"Multiple {name} blocks found")
    elif len(matches) == 1:
        content = "".join(
            line[2:] if line.startswith("# ") else line[1:]
            for line in matches[0].group("content").splitlines(keepends=True)
        )
        return tomlkit.parse(content)
    else:
        return None


def _parse_toml(fpath: str, standalone: bool = False) -> Optional[dict]:
    settings = {}
    with open(fpath, "r") as f:
        if standalone:
            settings = read_inline_metadata(f.read())
        else:
            settings = tomlkit.loads(f.read())
    return settings


def read_pymhf_settings(fpath: str, standalone: bool = False) -> dict:
    settings = _parse_toml(fpath, standalone)
    if not settings:
        return {}
    if standalone:
        return settings.get("tool", {}).get("pymhf", {})
    else:
        return settings.get("pymhf", {})


def write_pymhf_settings(settings: dict, fpath: str):
    """Write the pymhf settings to disk as a toml file.
    This will automatically add the `pymhf` top section.
    """
    with open(fpath, "w") as f:
        tomlkit.dump({"pymhf": settings}, f)



import ctypes
import re

import pytest
from typing_extensions import Self

from pymhf.core.functions import ArgData, FuncDef, _get_funcdef


def test_funcdef_flatten():
    """Test flattening arguments."""
    fd = FuncDef(
        ctypes.c_uint32,
        [ArgData("a", ctypes.c_uint32), ArgData("b", ctypes.c_uint64), ArgData("c", ctypes.c_float)],
    )
    assert fd.restype == ctypes.c_uint32
    assert fd.arg_names == ["a", "b", "c"]
    assert fd.arg_types == [ctypes.c_uint32, ctypes.c_uint64, ctypes.c_float]
    with pytest.raises(ValueError, match=re.escape("Missing argument(s): ['b', 'c']")):
        fd.flatten(1)
    with pytest.raises(ValueError, match=re.escape("Missing argument(s): ['c']")):
        fd.flatten(1, b=2)
    with pytest.raises(ValueError, match=re.escape("Missing argument(s): ['b']")):
        fd.flatten(1, c=2)
    assert fd.flatten(1, 2, 3) == [1, 2, 3]
    assert fd.flatten(a=1, b=3, c=5) == [1, 3, 5]


def test_funcdef_flatten_with_defaults():
    """Test flattening arguments when the FuncDef has default values."""
    fd = FuncDef(
        ctypes.c_uint32,
        [ArgData("a", ctypes.c_uint32), ArgData("b", ctypes.c_uint64), ArgData("c", ctypes.c_float)],
        {"b": 5, "c": 3.5},
    )
    assert fd.restype == ctypes.c_uint32
    assert fd.arg_names == ["a", "b", "c"]
    assert fd.arg_types == [ctypes.c_uint32, ctypes.c_uint64, ctypes.c_float]
    assert fd.flatten(1) == [1, 5, 3.5]
    assert fd.flatten(1, b=2) == [1, 2, 3.5]
    assert fd.flatten(1, c=2) == [1, 5, 2]
    assert fd.flatten(a=1, b=3, c=7) == [1, 3, 7]

    fd = FuncDef(
        ctypes.c_uint32,
        [ArgData("a", ctypes.c_uint32), ArgData("b", ctypes.c_uint64), ArgData("c", ctypes.c_float)],
        {"a": 99, "b": 5, "c": 3.5},
    )
    assert fd.flatten() == [99, 5, 3.5]


def test_get_funcdef_function():
    """Test getting the FuncDef for functions."""

    # Function with normal args and a None return.
    def func(x: ctypes.c_uint32, y: ctypes.c_uint64 = 1234) -> None:
        pass

    fd = _get_funcdef(func)
    assert fd.restype is None
    assert fd.arg_names == ["x", "y"]
    assert fd.arg_types == [ctypes.c_uint32, ctypes.c_uint64]
    assert fd.defaults == {"y": 1234}

    # Function with no args.
    def func2() -> ctypes.c_uint32:
        pass

    fd = _get_funcdef(func2)
    assert fd.restype == ctypes.c_uint32
    assert fd.arg_names == []
    assert fd.arg_types == []
    assert fd.defaults == {}

    class ID(ctypes.Structure):
        _fields_ = []

    # Function with an argument which is a pointer.
    def func3(id_: ctypes._Pointer[ID]) -> ctypes.c_bool:
        pass

    fd = _get_funcdef(func3)
    assert fd.restype == ctypes.c_bool
    assert fd.arg_names == ["id_"]
    assert fd.arg_types == [ctypes.POINTER(ID)]
    assert fd.defaults == {}

    # Function with mixed stringified and non-stringified args.
    # (emulates `from __future__ import annotations`)
    def func4(a: "ctypes.c_int32" = 42, b: ctypes.c_uint16 = 4):
        pass

    fd = _get_funcdef(func4)
    assert fd.restype is None
    assert fd.arg_names == ["a", "b"]
    assert fd.arg_types == [ctypes.c_int32, ctypes.c_uint16]
    assert fd.defaults == {"a": 42, "b": 4}

    # Function with an invalid argument types.
    def func5(a: int):
        pass

    with pytest.raises(TypeError, match=re.escape("Invalid type <class 'int'> for argument 'a'")):
        _get_funcdef(func5)

    def func6(a, b: ctypes.c_int64):
        pass

    with pytest.raises(
        TypeError,
        match=re.escape("The argument(s) ['a'] for 'func6' do not have type hints provided."),
    ):
        _get_funcdef(func6)

    def func7() -> int:
        pass

    with pytest.raises(
        TypeError,
        match="Return type must be a subclass of a ctypes.Structure or a simple ctypes type",
    ):
        _get_funcdef(func7)


def test_get_funcdef_method():
    """Test getting the FuncDef for methods."""

    class MyClass:
        def thing(self):
            # Very boring method with no args or return value.
            pass

        def thing2(self, x: ctypes.c_uint32, y: ctypes.c_uint16 = 7) -> ctypes.c_uint32:
            # Fairly boring method with some args and a default value.
            pass

        @staticmethod
        def thing3(x: ctypes.c_float = 999):
            # Static method.
            pass

        def thing4(self: Self, x: "ctypes.c_uint32"):
            # Valid "stringified" type.
            pass

        def thing5(self, x: "int"):
            # Invalid "stringified" type.
            pass

        def thing6(self, x: ctypes.c_uint32, y=7, z=None) -> ctypes.c_uint32:
            # Argument missing type hint.
            pass

        def thing7(self) -> "int":
            # Invalid "stringified" return type.
            pass

    fd = _get_funcdef(MyClass.thing)
    assert fd.restype is None
    assert fd.arg_names == []
    assert fd.arg_types == []
    assert fd.defaults == {}

    fd = _get_funcdef(MyClass.thing2)
    assert fd.restype == ctypes.c_uint32
    assert fd.arg_names == ["x", "y"]
    assert fd.arg_types == [ctypes.c_uint32, ctypes.c_uint16]
    assert fd.defaults == {"y": 7}

    fd = _get_funcdef(MyClass.thing3)
    assert fd.restype is None
    assert fd.arg_names == ["x"]
    assert fd.arg_types == [ctypes.c_float]
    assert fd.defaults == {"x": 999}

    fd = _get_funcdef(MyClass.thing4)
    assert fd.restype is None
    assert fd.arg_names == ["x"]
    assert fd.arg_types == [ctypes.c_uint32]
    assert fd.defaults == {}

    with pytest.raises(TypeError, match=re.escape("Invalid type <class 'int'> for argument 'x'")):
        fd = _get_funcdef(MyClass.thing5)

    with pytest.raises(
        TypeError,
        match=re.escape("The argument(s) ['y', 'z'] for 'thing6' do not have type hints provided."),
    ):
        fd = _get_funcdef(MyClass.thing6)

    with pytest.raises(
        TypeError,
        match="Return type must be a subclass of a ctypes.Structure or a simple ctypes type",
    ):
        fd = _get_funcdef(MyClass.thing7)



import ctypes
import logging
import sys
from gc import get_referents
from types import FunctionType, ModuleType
from typing import Iterable, Optional, Type, TypeVar, Union

import pymem
import pymem.pattern
import pymem.process
from pymem.ressources.structure import MODULEINFO

import pymhf.core._internal as _internal
import pymhf.core.caching as cache

__all__ = ["getsize", "get_addressof", "map_struct", "find_pattern_in_binary"]

# Custom objects know their class.
# Function objects seem to know way too much, including modules.
# Exclude modules as well.
BLACKLIST = type, ModuleType, FunctionType


mem_logger = logging.getLogger("MemUtils")

MEM_ACCESS_R = 0x100  # Read only.
MEM_ACCESS_RW = 0x200  # Read and Write access.


ctypes.pythonapi.PyMemoryView_FromMemory.argtypes = (
    ctypes.c_char_p,
    ctypes.c_ssize_t,
    ctypes.c_int,
)
ctypes.pythonapi.PyMemoryView_FromMemory.restype = ctypes.py_object


# TypeVar for the map_struct so that we can correctly get the returned type to
# be the same as the input type.
CTYPES = Union[ctypes._SimpleCData, ctypes.Structure, ctypes._Pointer]
Struct = TypeVar("Struct", bound=CTYPES)

# Temporary solution to avoid having to get the handles and suche every time we need to do a look up.
# "handle-module" cache
hm_cache: dict[str, tuple[int, MODULEINFO]] = {}
# Temporary solution to create a mapping of pattern/binary pairs to the offset within the binary.
offset_cache = {}


def getsize(obj):
    """Sum size of object & members."""
    if isinstance(obj, BLACKLIST):
        raise TypeError("getsize() does not take argument of type: " + str(type(obj)))
    seen_ids = set()
    size = 0
    objects = [obj]
    while objects:
        need_referents = []
        for _obj in objects:
            if not isinstance(_obj, BLACKLIST) and id(_obj) not in seen_ids:
                seen_ids.add(id(_obj))
                size += sys.getsizeof(_obj)
                need_referents.append(_obj)
        objects = get_referents(*need_referents)
    try:
        _len = len(obj)
    except TypeError:
        _len = None
    return size, _len


def chunks(lst: Iterable, n: int):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def match(patt: bytes, input: bytes):
    """Check whether or not the pattern matches the provided bytes."""
    for i, char in enumerate(patt):
        if not (char == b"." or char == input[i]):
            return False
    return True


def get_mem(offset: int, size: int) -> bytes:
    return (ctypes.c_char * size).from_address(offset)


def pprint_mem(offset: int, size: int, stride: Optional[int] = None) -> str:
    # TODO: Make this print a much nicer output... It sucks right now...
    if not offset:
        # If we are passed in an offset of 0, don't even try.
        return ""
    _data = (ctypes.c_char * size).from_address(offset)
    if stride:
        result = " ".join([f"{x:02X}".upper() for x in range(stride)]) + "\n"
        for chunk in chunks(_data, stride):
            result += " ".join([f"{k:02X}".upper() for k in chunk]) + "\n"
        return "\n" + result
    else:
        return "\n" + " ".join([k.hex().upper() for k in _data])


def _hex_repr(val: int, as_hex: bool) -> str:
    if as_hex:
        return hex(val)
    else:
        return str(val)


def get_field_info(obj, logger=None, indent: int = 0, as_hex: bool = True, max_depth: int = -1):
    if indent == max_depth:
        return
    if isinstance(obj, ctypes.Structure):
        # Need to get the actual class object to iterate over its' fields:
        cls_obj = obj.__class__
        has_values = True
    elif isinstance(obj, ctypes.Array):
        cls_obj = obj[0]
        has_values = True
    else:
        try:
            if issubclass(obj, ctypes.Structure):
                cls_obj = obj
                has_values = False
            elif issubclass(obj, ctypes.Array):
                cls_obj = obj._type_
                has_values = False
            else:
                raise TypeError(f"obj {obj} must be an instance of a ctypes.Structure or a subclass.")
        except TypeError as e:
            yield obj.__mro__
            raise TypeError(f"!!! obj {obj} must be an instance of a ctypes.Structure or a subclass.") from e
    for field, field_type in cls_obj._fields_:
        if has_values:
            val = getattr(obj, field)
            # if isinstance(val, ctypes.Array):
            #     val = [x for x in val]
        field_data: ctypes._CField = getattr(cls_obj, field)
        offset = _hex_repr(field_data.offset, as_hex)
        size = _hex_repr(field_data.size, as_hex)
        if has_values and not isinstance(val, ctypes.Structure):
            msg = f"{field} ({field_type.__name__}): size: {size} offset: {offset} value: {val}"
        else:
            msg = f"{field} ({field_type.__name__}): size: {size} offset: {offset}"
        msg = indent * "  " + msg
        yield msg
        if not issubclass(field_type, (ctypes._SimpleCData, ctypes.Array, ctypes._Pointer)):
            if has_values:
                for _msg in get_field_info(val, logger, indent + 1, as_hex, max_depth):
                    yield _msg
            else:
                for _msg in get_field_info(field_type, logger, indent + 1, as_hex, max_depth):
                    yield _msg


def get_addressof(obj: CTYPES) -> int:
    """Get the address in memory of some object.
    If obj is a pointer, this will return the address pointed to."""
    try:
        # If it's a pointer, this is the branch that is used.
        return ctypes.cast(obj, ctypes.c_void_p).value
    except Exception:
        # TODO: Get correct error type.
        # Otherwise fallback to the usual method.
        return ctypes.addressof(obj)


def _get_memview(offset: int, type_: Type[ctypes.Structure]) -> memoryview:
    """Return a memoryview which covers the region of memory specified by the
    struct provided.

    Parameters
    ----------
    offset:
        The memory address to start reading the struct from.
    type_:
        The type of the ctypes.Structure to be loaded at this location.
    """
    return ctypes.pythonapi.PyMemoryView_FromMemory(
        ctypes.cast(offset, ctypes.c_char_p),
        ctypes.sizeof(type_),
        MEM_ACCESS_RW,
    )


def _get_memview_with_size(offset: int, size: int) -> Optional[memoryview]:
    """Return a memoryview which covers the region of memory specified by the
    struct provided.

    Parameters
    ----------
    offset:
        The memory address to start reading the struct from.
    type_:
        The type of the ctypes.Structure to be loaded at this location.
    """
    if not offset:
        return None
    return ctypes.pythonapi.PyMemoryView_FromMemory(
        ctypes.cast(offset, ctypes.c_char_p),
        size,
        MEM_ACCESS_RW,
    )


def map_struct(offset: int, type_: Type[Struct]) -> Struct:
    """Return an instance of the `type_` struct provided which shares memory
    with the provided offset.
    Note that the amount of memory to read is automatically determined by the
    size of the struct provided.

    Parameters
    ----------
    offset:
        The memory address to start reading the struct from.
    type\_:
        The type of the ctypes.Structure to be loaded at this location.

    Returns
    -------
    An instance of the input type.
    """
    if not offset:
        raise ValueError("Offset is 0. This would result in a segfault or similar")
    instance = ctypes.cast(offset, ctypes.POINTER(type_))
    return instance.contents


def pattern_to_bytes(patt: str) -> bytes:
    """Take a pattern that looks like `8C 14 23 56 ?? 12` (etc) and convert it to a bytes object which can
    be searched with pymem.
    The format is what is provided by the IDA plugin `SigMakerEx` and the `??` values indicate a wildcard.
    This also supports the format by the IDA plugin `IDA-Fusion` which produces single `?` values for
    wildcards.
    """
    split = patt.split(" ")
    return b"".join([f"\\x{x}".encode() if (x != "??" and x != "?") else b"." for x in split])


def _get_binary_info(binary: str) -> Optional[tuple[int, MODULEINFO]]:
    if binary not in hm_cache:
        try:
            pm_process = pymem.Pymem(_internal.EXE_NAME)
            handle = pm_process.process_handle
            if (module := cache.module_map.get(binary)) is None:
                return None
            hm_cache[binary] = (handle, module)
            return (handle, module)
        except TypeError:
            return None
    else:
        return hm_cache[binary]


def find_pattern_in_binary(
    pattern: str,
    return_multiple: bool = False,
    binary: Optional[str] = None,
) -> Union[int, list[int], None]:
    """
    Find a pattern in the specified binary. This is for the most part a wrapper around pymem's
    pattern_scan_module function so that we can just call this with a module name (often the name of the
    binary we ran, but it could be some other dll under the same process.)
    """
    if binary is None:
        binary = _internal.EXE_NAME
    hm = _get_binary_info(binary)
    if not hm:
        return None
    handle, module = hm
    # Create a key which is the original pattern and the binary so that we may cache the result.
    key = (pattern, binary)
    if (_offset := offset_cache.get(key)) is not None:
        return _offset
    patt = pattern_to_bytes(pattern)
    _offset = pymem.pattern.pattern_scan_module(handle, module, patt, return_multiple=return_multiple)
    if _offset:
        _offset = _offset - module.lpBaseOfDll
    # Cache even if there is no result (so we don't repeatedly look for it when it's not there in case there
    # is an issue.)
    offset_cache[key] = _offset
    return _offset



import os
import os.path as op
import re
from typing import Optional

PATH_RE = re.compile(r"^\{(?P<tag>EXE_DIR|USER_DIR|CURR_DIR)\}(?P<rest>[^{}]*)$")


def canonicalize_setting(
    value: Optional[str],
    plugin_name: Optional[str],
    module_dir: str,
    exe_dir: Optional[str] = None,
    suffix: Optional[str] = None,
) -> Optional[str]:
    """Convert the "magic" names into real values.

    Possible keys:
    - EXE_DIR
    - USER_DIR / "~"
    - CURR_DIR / "."
    """

    # This can receive None as the value.
    # In this case we simply return as we don't want to do anything with it.
    if value is None:
        return None

    # Parse the value to determine what directory type we are asking for.
    tag = None
    rest = tuple()
    if (m := re.match(PATH_RE, value)) is not None:
        tag = m["tag"]
        if m["rest"]:
            rest = op.split(m["rest"].strip("/\\"))
    else:
        if value == ".":
            tag = "CURR_DIR"
        elif value == "~":
            tag = "USER_DIR"
        elif not op.exists(value):
            raise ValueError(f"Path doesn't exist: {value}")

    # If the path provided already exists, simply return it.
    if tag is None and op.exists(value):
        return value

    if suffix is None:
        _suffix = rest
    else:
        _suffix = rest + (suffix,)

    if tag == "USER_DIR":
        appdata_data = os.environ.get("APPDATA", op.expanduser("~"))
        if appdata_data == "~":
            # In this case the APPDATA environment variable isn't set and ~ also fails to resolve.
            # Raise a error and stop.
            print("Critical Error: Cannot find user directory. Ensure APPDATA environment variable is set")
            exit()
        if plugin_name is not None:
            return op.realpath(op.join(appdata_data, "pymhf", plugin_name, *_suffix))
        else:
            raise ValueError("{USER_DIR} cannot be used for single-file mods.")
    elif tag == "EXE_DIR":
        if exe_dir:
            return op.realpath(op.join(exe_dir, *_suffix))
        else:
            raise ValueError("Exe directory cannot be determined")
    elif tag == "CURR_DIR":
        return op.realpath(op.join(module_dir, *_suffix))



import importlib
import importlib.util
import logging
import os.path as op
import string
import sys
import traceback
from types import ModuleType
from typing import Optional

logger = logging.getLogger("pymfh.core.importing")


VALID_CHARS = string.ascii_letters + string.digits + "_"


def _clean_name(name: str) -> str:
    """Remove any disallowed characters from the filename so that we get a
    valid module name.
    """
    out = ""
    for char in name:
        if char not in VALID_CHARS:
            out += "_"
        else:
            out += char
    return out


def import_file(fpath: str) -> Optional[ModuleType]:
    try:
        module_name = _clean_name(op.splitext(op.basename(fpath))[0])
        if op.isdir(fpath):
            # If a directory is passed in, then add __init__.py to it so that
            # we may correctly import it.
            fpath = op.join(fpath, "__init__.py")
        if spec := importlib.util.spec_from_file_location(module_name, fpath):
            module = importlib.util.module_from_spec(spec)
            module.__name__ = module_name
            module.__spec__ = spec
            sys.modules[module_name] = module
            spec.loader.exec_module(module)
            return module
        else:
            print("failed")
    except Exception:
        logger.error(f"Error loading {fpath}")
        logger.error(traceback.format_exc())



from enum import Enum
from typing import Any, NamedTuple, Optional, Protocol


# TODO: Fully deprecate.
class FUNCDEF(NamedTuple):
    restype: Any
    argtypes: list


class DetourTime(Enum):
    NONE = 0
    BEFORE = 1
    AFTER = 2


class KeyPressProtocol(Protocol):
    _hotkey: str
    _hotkey_press: str


class HookProtocol(Protocol):
    _disabled: bool
    _is_funchook: bool
    _is_manual_hook: bool
    _is_imported_func_hook: bool
    _is_exported_func_hook: bool
    _has__result_: bool
    _hook_func_name: str
    _hook_func_def: FUNCDEF
    _hook_offset: Optional[int]
    _hook_pattern: Optional[str]
    _hook_time: DetourTime
    _custom_trigger: Optional[str]
    _func_overload: Optional[str]
    _get_caller: Optional[bool]
    _noop: Optional[bool]
    _dll_name: str

    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...


class CallerHookProtocol(HookProtocol):
    def caller_address(self) -> int:
        """The address relative to the base of the calling binary that this functiuon detour was called from.
        Note that this will be the address one instruction after the caller since this value is where
        execution returns to after the detours and original function have been run.
        """
        ...


# TODO: Remove once we have got the binary able to be specified for function_hook decorators.
class ManualHookProtocol(HookProtocol):
    _hook_binary: Optional[str]



import ctypes
import inspect
from dataclasses import dataclass
from typing import Optional, Type, TypeVar, Union, _AnnotatedAlias, get_args

from typing_extensions import get_type_hints

from pymhf.extensions.ctypes import c_enum32

_T = TypeVar("_T", bound=Type[ctypes.Structure])

CTYPES = Union[
    ctypes._SimpleCData,
    ctypes.Structure,
    ctypes._Pointer,
    ctypes._Pointer_orig,  # The original, un-monkeypatched ctypes._Pointer object
    ctypes.Array,
    ctypes.Union,
    c_enum32,
]


@dataclass
class Field:
    datatype: CTYPES
    offset: Optional[int] = None


def partial_struct(cls: _T) -> _T:
    """Mark the decorated class as a partial struct.
    This will automatically construct the ``_field_`` attribute for this class
    """
    # Always get the calling frame and add the locals so that if we have any annotated fields they won't fail.
    # In the case of a struct which has no annotations this will be ever so slightly slower than if we did a
    # try... except.
    # However in the case of annotated fields, it's much faster, so it's beneficial to always do this.
    calling_frame = None
    if (cframe := inspect.currentframe()) is not None:
        calling_frame = cframe.f_back
    if calling_frame is not None:
        locals().update(calling_frame.f_locals)
    _fields_ = []
    curr_position = 0
    total_size = getattr(cls, "_total_size_", 0)
    # If there are no annotations, it's just an empty ctypes.Structure class.
    if not hasattr(cls, "__annotations__"):
        cls._fields_ = _fields_
        return cls
    # If there are, loop over the annotations and extract the info we need to construct the _fields_.
    for field_name, annotation in get_type_hints(cls, include_extras=True, localns=locals()).items():
        if not isinstance(annotation, _AnnotatedAlias):
            # In this case it's just the type.
            field_type = annotation
            field_offset = None
        else:
            # If the annotation is a Field object, get info from it, otherwise it must be an integer
            # specifying the offset.
            metadata = annotation.__metadata__
            if len(metadata) != 1:
                raise ValueError(f"The field {field_name!r} has an invalid annotation: {annotation}")
            metadata = metadata[0]
            if isinstance(metadata, Field):
                field_type = metadata.datatype
                field_offset = metadata.offset
            else:
                field_type = annotation.__origin__
                field_offset = metadata
        if not issubclass(field_type, get_args(CTYPES)):
            raise ValueError(f"The field {field_name!r} has an invalid type: {field_type}")
        if field_offset is not None and not isinstance(field_offset, int):
            raise ValueError(f"The field {field_name!r} has an invalid offset: {field_offset!r}")
        if field_offset and field_offset > curr_position:
            padding_bytes = field_offset - curr_position
            _fields_.append((f"_padding_{curr_position:X}", ctypes.c_ubyte * padding_bytes))
            curr_position += padding_bytes
        field_alignment = ctypes.alignment(field_type)
        if curr_position % field_alignment != 0:
            # If the field is not aligned to the correct position, then move the current position forward
            # to ensure it's right.
            # Don't bother adding this as padding since it will just add extra unnecessary fields.
            diff = field_alignment - (curr_position % field_alignment)
            curr_position += diff
        _fields_.append((field_name, field_type))
        curr_position += ctypes.sizeof(field_type)
    if total_size and curr_position < total_size:
        padding_bytes = total_size - curr_position
        _fields_.append((f"_padding_{curr_position:X}", ctypes.c_ubyte * padding_bytes))
    cls._fields_ = _fields_
    return cls



import ast
import ctypes
import inspect
import logging
import struct
import traceback
from collections import defaultdict
from collections.abc import Callable
from ctypes import CFUNCTYPE
from typing import Any, Optional, Type, Union, cast

import cyminhook
from typing_extensions import Concatenate, Generic, ParamSpec, Self, TypeVar, deprecated

import pymhf.core._internal as _internal
from pymhf.core._types import (
    FUNCDEF,
    CallerHookProtocol,
    DetourTime,
    HookProtocol,
    KeyPressProtocol,
    ManualHookProtocol,
)
from pymhf.core.functions import FuncDef, _get_funcdef
from pymhf.core.memutils import _get_binary_info, find_pattern_in_binary, get_addressof
from pymhf.core.module_data import module_data
from pymhf.utils.iced import HAS_ICED, generate_load_stack_pointer_bytes, get_first_jmp_addr

hook_logger = logging.getLogger("HookManager")


BITS = struct.calcsize("P") * 8


# Currently unused, but can maybe figure out how to utilise it.
# It currently doesn't work I think because we subclass from the cyminhook class
# which is cdef'd, and I guess ast falls over trying to get the actual source...
# Can possible use annotations and inspect the return type (return `None`
# explictly eg.) to give some hints. Maybe just raise warnings etc.
def _detour_is_valid(f):
    for node in ast.walk(ast.parse(inspect.getsource(f))):
        if isinstance(node, ast.Return):
            return True
    return False


# A VERY rudimentary cache untill we implement gh-2.
pattern_cache: dict[str, int] = {}

_FunctionHook_overloads: dict = defaultdict(lambda: dict())


# TODO: Move to a different file with `Mod` from mod_loader.py
class FuncHook(cyminhook.MinHook):
    original: Callable[..., Any]
    target: int
    detour: Callable[..., Any]
    signature: Callable[..., Any]
    _name: str
    _invalid: bool = False
    _func_def: Optional[FUNCDEF]
    _offset_is_absolute: bool

    def __init__(
        self,
        detour_name: str,
        *,
        offset: Optional[int] = None,
        pattern: Optional[str] = None,
        func_def: Optional[FUNCDEF] = None,
        overload: Optional[str] = None,
        binary: Optional[str] = None,
        offset_is_absolute: bool = False,
    ):
        self._offset_is_absolute = offset_is_absolute
        if self._offset_is_absolute and offset is not None:
            self.target = offset
            self._offset = None
        else:
            self._offset = offset
        self._pattern = pattern
        self._func_def = func_def
        self._binary = binary
        if BITS == 64:
            self._rsp_addr = ctypes.c_ulonglong(0)
        else:
            self._rsp_addr = ctypes.c_ulong(0)
        if self._binary is not None:
            if (hm := _get_binary_info(self._binary)) is not None:
                _, module = hm
                self._binary_base = module.lpBaseOfDll
        else:
            self._binary = _internal.EXE_NAME
            self._binary_base = _internal.BASE_ADDRESS

        self._has_noop = False
        self._before_detours: list[HookProtocol] = []
        self._after_detours: list[HookProtocol] = []
        self._after_detours_with_results: list[HookProtocol] = []
        # Disabled detours will go here. This will include hooks disabled by the
        # @disable decorator, as well as hooks which are one-shots and have been
        # run.
        self._disabled_detours: set[HookProtocol] = set()
        self._oneshot_detours: dict[HookProtocol, HookProtocol] = {}
        self.overload = overload
        self.state = None
        self._name = detour_name
        self._initialised = False

    @property
    def caller_address(self):
        if self._rsp_addr.value:
            return self._rsp_addr.value - _internal.BASE_ADDRESS
        return 0

    @property
    def name(self):
        if self.overload is not None:
            return f"{self._name}({self.overload})"
        else:
            return self._name

    def _init(self):
        """Actually initialise all the data. This is defined separately so that any function which is marked
        with @disable doesn't get initialised.
        """
        offset = None
        # 1. Check to see if an offset is provided. If so, use this as the offset to find the address at.
        if self._offset is not None:
            self.target = self._binary_base + self._offset

        # 2. If there is no offset provided, check to see if a pattern is provided. If so use this to find
        #    the offset.
        elif self._pattern is not None:
            # Lookup the pattern in the pattern cache. If that fails find it in the binary.
            if (offset := pattern_cache.get(self._pattern)) is None:
                offset = find_pattern_in_binary(self._pattern, False, self._binary)
            if offset is not None:
                self._offset = offset
                pattern_cache[self._pattern] = offset
                self.target = self._binary_base + offset
                hook_logger.debug(f"Found {self._pattern} at 0x{offset:X}")
            else:
                hook_logger.error(
                    f"Could not find pattern {self._pattern}... Hook {self._name!r} won't be added"
                )
                self._invalid = True
                return

        # 3. If there is still no offset, look up the pattern in the module_data to get offset
        elif (_pattern := module_data.FUNC_PATTERNS.get(self._name)) is not None:
            if isinstance(_pattern, str):
                if (offset := pattern_cache.get(_pattern)) is None:
                    offset = find_pattern_in_binary(_pattern, False, self._binary)
            else:
                if (overload_pattern := _pattern.get(self.overload)) is not None:
                    _pattern = overload_pattern
                    if overload_pattern in pattern_cache:
                        offset = pattern_cache[overload_pattern]
                    else:
                        offset = find_pattern_in_binary(overload_pattern, False, self._binary)
                else:
                    first = list(_pattern.items())[0]
                    _pattern = first
                    hook_logger.warning(f"No overload was provided for {self._name}. ")
                    hook_logger.warning(f"Falling back to the first overload ({first[0]})")
                    offset = find_pattern_in_binary(first[1], False, self._binary)
            if offset is not None:
                self._offset = offset
                pattern_cache[_pattern] = offset
                self.target = self._binary_base + offset
                hook_logger.debug(f"Found {self.name} with pattern {_pattern!r} at 0x{offset:X}")

        # 4. If there is still no offset, look up the offset in the module_data.
        # Note: If this is created by passing in an absolute offset, then the above 3 conditions will fail,
        # but this one will not as self.target will have already been assigned.
        if not self.target:
            _offset = module_data.FUNC_OFFSETS.get(self._name)
            if _offset is not None:
                if isinstance(_offset, int):
                    self.target = self._binary_base + _offset
                else:
                    # This is an overload
                    if (overload_offset := _offset.get(self.overload)) is not None:
                        self.target = self._binary_base + overload_offset
                    else:
                        # Need to fallback on something. Raise a warning that no
                        # overload was defined and that it will fallback to the
                        # first entry in the dict.
                        first = list(_offset.items())[0]
                        hook_logger.warning(f"No overload was provided for {self._name}. ")
                        hook_logger.warning(f"Falling back to the first overload ({first[0]})")
                        self.target = self._binary_base + first[1]
            else:
                hook_logger.error(f"Cannot find the function {self._name} in {_internal.EXE_NAME}")
                self._invalid = True
                return

        # 5. if func_sig is provided, use it, otherwise look it up.
        if self._func_def is not None:
            self.signature = CFUNCTYPE(self._func_def.restype, *self._func_def.argtypes)
        else:
            if (sig := module_data.FUNC_CALL_SIGS.get(self._name)) is not None:
                if isinstance(sig, FUNCDEF):
                    self.signature = CFUNCTYPE(sig.restype, *sig.argtypes)
                    hook_logger.debug(
                        f"Function {self._name} return type: {sig.restype} args: {sig.argtypes}"
                    )
                    if self.overload is not None:
                        hook_logger.warning(
                            f"An overload was provided for {self._name} but no overloaded"
                            " function definitions exist. This function may fail."
                        )
                else:
                    # Look up the overload:
                    if (osig := sig.get(self.overload)) is not None:  # type: ignore
                        self.signature = CFUNCTYPE(osig.restype, *osig.argtypes)
                        hook_logger.debug(
                            f"Function {self._name} return type: {osig.restype} args: {osig.argtypes}"
                        )
                    else:
                        # Need to fallback on something. Raise a warning that no overload was defined and that
                        # it will fallback to the first entry in the dict.
                        first = list(sig.items())[0]
                        hook_logger.warning(f"No function arguments overload was provided for {self._name}. ")
                        hook_logger.warning(f"Falling back to the first overload ({first[0]})")
                        self.signature = CFUNCTYPE(first[1].restype, *first[1].argtypes)
            else:
                hook_logger.error(f"{self._name} has no known call signature")
                self._invalid = True
                return

        self._initialised = True

    def _determine_detour_list(self, detour: HookProtocol) -> Optional[list[HookProtocol]]:
        # Determine when the hook should be run. Don't add the detour yet
        # because if the hook is a one-shot then we need to know when to run it.
        detour_list = None
        if detour._hook_time == DetourTime.BEFORE:
            detour_list = self._before_detours
        elif detour._hook_time == DetourTime.AFTER:
            # Check to see if the detour has the `_result_` argument.
            # This will have been determined already when the function was
            # decorated
            # TODO: Ensure this works if there is some other funky decoration
            # shenanigans (eg. with a one-shot decorator...)
            if getattr(detour, "_has__result_", False):
                detour_list = self._after_detours_with_results
            else:
                detour_list = self._after_detours
        else:
            hook_logger.error(f"Detour {detour} has an invalid detour time: {detour._hook_time}")
        return detour_list

    def add_detour(self, detour: HookProtocol):
        """Add the provided detour to this FuncHook."""
        # If the hook has the `_disabled` attribute, then don't add the detour.
        if getattr(detour, "_disabled", False):
            self._disabled_detours.add(detour)
            return

        if getattr(detour, "_noop", False):
            self._has_noop = True
            hook_logger.warning(
                f"The hook {detour._hook_func_name} has been marked as NOOP. If there are multiple detours "
                "registered for this hook they may behave weirdly and the game may crash if this hook is not "
                "defined well."
            )

        # Determine the detour list to use. If none, then return.
        if (detour_list := self._determine_detour_list(detour)) is None:
            hook_logger.error(
                f"Unable to assign {detour} to a detour type. Please check it. It has not been added."
            )
            return

        if not getattr(detour, "_is_one_shot", False):
            # If we aren't a one-shot detour, then add it to the list.
            detour_list.append(detour)

        # If the hook is a one-shot, wrap it so that it can remove itself once
        # it's executed.
        if getattr(detour, "_is_one_shot", False):

            def _one_shot(
                *args,
                detour: HookProtocol = detour,
                detour_list: list[HookProtocol] = detour_list,
            ):
                # NOTE: This may not work well if the code is called from multiple threads at the same time.
                try:
                    detour(*args)
                    self._disabled_detours.add(detour)
                    detour_list.remove(self._oneshot_detours[detour])
                except ValueError:
                    hook_logger.warning(
                        f"Had an issue removing one-shot {detour._hook_func_name} from detour list."
                    )
                except Exception:
                    hook_logger.error(traceback.format_exc())

            self._oneshot_detours[detour] = _one_shot
            detour_list.append(_one_shot)

        # If the detour needs the `caller_address` property, add it.
        if getattr(detour, "_get_caller", False) is True:
            # We need to get the type of the class and assign the attribute to the class function itself.
            detour.__func__.caller_address = lambda: self.caller_address

    def remove_detour(self, detour: HookProtocol):
        """Remove the provided detour from this FuncHook."""
        # Determine the detour list to use. If none, then return.
        if (detour_list := self._determine_detour_list(detour)) is None:
            return

        if detour in detour_list:
            detour_list.remove(detour)
        # Try and remove the hook from the diabled list also in case it's there.
        if detour in self._disabled_detours:
            self._disabled_detours.remove(detour)
        # Also check for one-shot detours. They may or may not have been called,
        # but we don't really care.
        # If it has been called then it will be in the disabled detours and so
        # we will have handled it already.
        # If it hasn't then we will look up the mapping now and remove it.
        if detour in self._oneshot_detours:
            one_shot_detour = self._oneshot_detours.pop(detour)
            if one_shot_detour in detour_list:
                detour_list.remove(one_shot_detour)

        # If we have no more detours remaining, then we disable and close this hook so that we may free it to
        # allow us to correctly re-create it later if need be.
        if not detour_list:
            self.disable()
            self.close()

        # Check to see if we have any detours marked as NOOP. If not, make sure the flag is set to false.
        # We can always just check the before detours since only a before detour can be marked as NOOP.
        self._has_noop == any([getattr(d, "_noop", False) for d in self._before_detours])

    @property
    def _should_enable(self):
        return (
            len(self._before_detours) > 0
            or len(self._after_detours) > 0
            or len(self._after_detours_with_results)
        )

    def bind(self) -> bool:
        """Actually initialise the base class. Returns whether the hook is bound."""
        if not self._should_enable or self._invalid:
            return False

        self.detour = self._compound_detour

        try:
            super().__init__(signature=self.signature, target=self.target)
        except cyminhook._cyminhook.Error as e:  # type: ignore
            if e.status == cyminhook._cyminhook.Status.MH_ERROR_ALREADY_CREATED:
                hook_logger.info(
                    "Hook is already created. This shouldn't be possible. Please raise an issue on github..."
                )
            hook_logger.error(f"Failed to initialize hook {self._name} at 0x{self.target:X}")
            hook_logger.error(e)
            hook_logger.error(e.status.name[3:].replace("_", " "))
            self.state = "failed"
            return False
        self.state = "initialized"
        return True

    def _compound_detour(self, *args):
        ret = None

        # Loop over the before detours, keeping the last none-None return value.
        try:
            for i, func in enumerate(self._before_detours):
                r = func(*args)
                if r is not None:
                    ret = r
        except Exception:
            bad_detour = self._before_detours.pop(i)
            hook_logger.error(f"There was an error with detour {bad_detour}. It has been disabled.")
            hook_logger.error(traceback.format_exc())
            self._disabled_detours.add(bad_detour)

        # If we don't have any decorators which NOOP the original function then run as usual.
        if not self._has_noop:
            # If we get a return value that is not None, then pass it through.
            if ret is not None:
                result = self.original(*ret)
            else:
                result = self.original(*args)
            after_ret = None
        # If we have any NOOP's, then we don't want to run the original, and instead will have the last result
        # returned from our functions as the "result".
        else:
            result = ret
            after_ret = ret

        # Now loop over the after functions. We'll need to handle the cases of
        # functions which take the `_result_` kwarg, and those that don't.
        try:
            for i, func in enumerate(self._after_detours):
                after_ret = func(*args)
            i = None
            for j, func in enumerate(self._after_detours_with_results):
                after_ret = func(*args, _result_=result)
            j = None
        except Exception:
            if i is not None:
                bad_detour = self._after_detours.pop(i)
            else:
                bad_detour = self._after_detours_with_results.pop(j)
            hook_logger.error(f"There was an error with detour {bad_detour}. It has been disabled.")
            hook_logger.error(traceback.format_exc())
            self._disabled_detours.add(bad_detour)

        if after_ret is not None:
            return after_ret
        return result

    def close(self):
        super().close()
        self.state = "closed"

    def enable(self):
        if self._should_enable:
            super().enable()
            self.state = "enabled"

    def disable(self):
        if self.state == "enabled":
            super().disable()
            self.state = "disabled"

    @property
    def offset(self) -> int:
        """The relative offset of the target to the binary base."""
        return self.target - self._binary_base


@deprecated("Use @function_hook(signature=...) instead", category=DeprecationWarning)
class HookFactory:
    _name: str
    _templates: Optional[tuple[str]] = None
    _overload: Optional[str] = None

    @classmethod
    def overload(cls, overload_args):
        # TODO: Improve type hinting and possible make this have a generic arg
        # type to simplify the logic...
        raise NotImplementedError

    @staticmethod
    def _set_detour_as_funchook(
        detour: Callable[..., Any],
        cls: Optional["HookFactory"] = None,
        detour_name: Optional[str] = None,
    ):
        """Set all the standard attributes required for a function hook."""
        setattr(detour, "_is_funchook", True)
        setattr(detour, "_is_manual_hook", False)
        setattr(detour, "_is_imported_func_hook", False)
        setattr(detour, "_is_exported_func_hook", False)
        setattr(detour, "_has__result_", False)
        setattr(detour, "_noop", False)
        if cls:
            if hasattr(cls, "_overload"):
                setattr(detour, "_func_overload", cls._overload)
            else:
                setattr(detour, "_func_overload", None)
            if hasattr(cls, "_name"):
                setattr(detour, "_hook_func_name", cls._name)
        else:
            if detour_name is None:
                raise ValueError("class used as detour must have a name")
            else:
                setattr(detour, "_hook_func_name", detour_name)

    @classmethod
    def after(cls, detour: Callable[..., Any]) -> HookProtocol:
        """
        Run the detour *after* the original function.
        An optional `_result_` argument can be added as the final argument.
        If this argument is provided it will be the result of calling the original function.
        """
        HookFactory._set_detour_as_funchook(detour, cls)
        setattr(detour, "_hook_time", DetourTime.AFTER)
        if "_result_" in inspect.signature(detour).parameters.keys():
            setattr(detour, "_has__result_", True)
        return detour

    @classmethod
    def before(cls, detour: Callable[..., Any]) -> HookProtocol:
        """
        Run the detour *before* the original function.
        If this detour returns any values they must be the same types and order as the original arguments to
        the function. If this happens these values will be passed into the original function instead of the
        original arguments.
        """
        HookFactory._set_detour_as_funchook(detour, cls)
        setattr(detour, "_hook_time", DetourTime.BEFORE)
        return detour


@deprecated(
    "Use @function_hook(signature=...) or @function_hook(offset=...) instead",
    category=DeprecationWarning,
)
def manual_hook(
    name: str,
    offset: Optional[int] = None,
    pattern: Optional[str] = None,
    func_def: Optional[FUNCDEF] = None,
    detour_time: str = "before",
    binary: Optional[str] = None,
):
    """Manually hook a function.

    Parameters
    ----------
    name:
        The name of the function to hook. This doesn't need to be known, but any two manual hooks sharing the
        same name will be combined together so one should remember to keep the name/offset combination unique.
    offset:
        The offset in bytes relative to the start of the binary.
        To determine this, you normally subtract off the exe Imagebase value from the address in IDA (or
        similar program.)
    pattern:
        A pattern which can be used to unqiuely find the function to be hooked within the binary.
        The pattern must have a format like "01 23 45 67 89 AB CD EF".
        The format is what is provided by the IDA plugin `SigMakerEx` and the `??` values indicate a wildcard.
    func_def:
        The function arguments and return value. This is provided as a `pymhf.FUNCDEF` object.
        This argument is only optional if another function with the same offset and name has already been
        hooked in the same mod.
    detour_time:
        When the detour should run ("before" or "after")
    binary:
        If provided, this will be the name of the binary which the function being hooked is within.
        `offset` and `pattern` are found relative to/within the memory region of this binary.
    """

    def inner(detour: Callable[..., Any]) -> ManualHookProtocol:
        # if offset is None and pattern is None:
        #     raise ValueError(f"One of pattern or offset must be set for the manual hook: {detour}")
        HookFactory._set_detour_as_funchook(detour, None, name)
        if detour_time == "before":
            setattr(detour, "_hook_time", DetourTime.BEFORE)
        elif detour_time == "after":
            setattr(detour, "_hook_time", DetourTime.AFTER)
        # Set some manual values which can be retrieved when the func hook gets parsed.
        setattr(detour, "_is_manual_hook", True)
        setattr(detour, "_hook_offset", offset)
        setattr(detour, "_hook_pattern", pattern)
        setattr(detour, "_hook_binary", binary)
        setattr(detour, "_hook_func_def", func_def)
        if "_result_" in inspect.signature(detour).parameters.keys():
            setattr(detour, "_has__result_", True)
        return detour

    return inner


def NOOP(detour: HookProtocol) -> HookProtocol:
    """
    Specify the hook to not run the original function.
    This decorator must be used with extreme caution as it can cause the hooked program to not run correctly
    if not used right.
    The decorated function MUST return something of the same type as the original function if the hooked
    function normally returns something otherwise the hooked program will almost certainly crash.
    """
    if getattr(detour, "_hook_time", None) == DetourTime.BEFORE:
        setattr(detour, "_noop", True)
    else:
        hook_logger.warning(
            "NOOP decorator can only be applied to 'before' hooks."
            "Either change the detour to a before detour, or, if it is, ensure that this decorator is "
            "applied above the hook decorator."
        )
    return detour


def disable(obj):
    """
    Disable the current function or class.
    """
    setattr(obj, "_disabled", True)
    return obj


@deprecated("Use @function_hook(imported_name=...) instead", category=DeprecationWarning)
def imported(dll_name: str, func_name: str, func_def: FUNCDEF, detour_time: str = "after"):
    """Hook an imported function in `dll_name` dll.

    Parameters
    ----------
    dll_name:
        The name of the dll which contains the function.
    func_name:
        The name of the function in the dll which is to be hooked.
    func_def:
        The function signature.
    detour_time:
        Whether to run the detour before or after the original function.
    """

    def inner(detour: Callable[..., Any]) -> HookProtocol:
        HookFactory._set_detour_as_funchook(detour, None, func_name)
        setattr(detour, "_dll_name", dll_name)
        setattr(detour, "_is_imported_func_hook", True)
        setattr(detour, "_hook_func_def", func_def)
        if detour_time == "before":
            setattr(detour, "_hook_time", DetourTime.BEFORE)
        else:
            setattr(detour, "_hook_time", DetourTime.AFTER)
            if "_result_" in inspect.signature(detour).parameters.keys():
                setattr(detour, "_has__result_", True)
        return detour

    return inner


@deprecated("Use @function_hook(exported_name=...) instead", category=DeprecationWarning)
def exported(func_name: str, func_def: FUNCDEF, detour_time: str = "after"):
    """Hook an exported function.

    Parameters
    ----------
    func_name:
        The name of the function which is to be hooked.

        .. note::
            It is recommended that the function name is the "mangled" version.
            Ie. do not "demangle" the function name.
    func_def:
        The function signature.
    detour_time:
        Whether to run the detour before or after the original function.
    """

    def inner(detour: Callable[..., Any]) -> HookProtocol:
        HookFactory._set_detour_as_funchook(detour, None, func_name)
        setattr(detour, "_is_exported_func_hook", True)
        setattr(detour, "_hook_func_def", func_def)
        if detour_time == "before":
            setattr(detour, "_hook_time", DetourTime.BEFORE)
        else:
            setattr(detour, "_hook_time", DetourTime.AFTER)
        return detour

    return inner


def one_shot(func: HookProtocol) -> HookProtocol:
    """Run this detour once only."""
    setattr(func, "_is_one_shot", True)
    return func


def get_caller(func: HookProtocol) -> CallerHookProtocol:
    """Capture the address this hooked function was called from.
    This address will be acessible by the `caller_address()` method which will belong to the function that is
    decorated by this.

    Examples
    --------
    .. code:: py

        @get_caller
        @manual_hook("test_function", 0x12345678, FUNCDEF(restype=ctypes.void, argtypes=[ctypes.c_ulonglong]))
        def something(self, *args):
            logger.info(f"'test_function' called with {args} from 0x{self.something.caller_address():X}")
    """
    func._get_caller = True
    return func  # type: ignore


def on_key_pressed(event: str):
    """Register the provided event as a key press handler.
    When the key is pressed, the decorated function will be called.

    Parameters
    ----------
    event:
        The string representing the key which is to trigger the event.
    """

    def wrapped(func: Callable[..., Any]) -> KeyPressProtocol:
        setattr(func, "_hotkey", event)
        setattr(func, "_hotkey_press", "down")
        return func  # type: ignore

    return wrapped


def on_key_release(event: str):
    """Register the provided event as a key release handler.
    When the key is released, the decorated function will be called.

    Parameters
    ----------
    event:
        The string representing the key which is to trigger the event.
    """

    def wrapped(func: Callable[..., Any]) -> KeyPressProtocol:
        setattr(func, "_hotkey", event)
        setattr(func, "_hotkey_press", "up")
        return func  # type: ignore

    return wrapped


class HookManager:
    def __init__(self):
        self.hooks: dict[str, FuncHook] = {}
        # Keep a mapping of any hooks that try to be registered but fail.
        # These hooks will not be instances of classes, but the class type.
        self.failed_hooks: dict[str, Type[FuncHook]] = {}
        # A mapping of the custom event hooks which can be registered by modules
        # for individual mods.
        self.custom_callbacks: dict[str, dict[DetourTime, set[HookProtocol]]] = {}
        self._uninitialized_hooks: set[str] = set()

        self._get_caller_detours: set[str] = set()

    def _resolve_dependencies(self):
        """Resolve dependencies of hooks.
        This will get all the functions which are to be hooked and construct
        compound hooks as required.
        """
        # TODO: Make work.
        pass

    def _add_custom_callbacks(self, callbacks: set[HookProtocol]):
        """Add the provided function to the specified callback type."""
        for cb in callbacks:
            if (cb_type := cb._custom_trigger) is None:
                continue
            if cb_type not in self.custom_callbacks:
                self.custom_callbacks[cb_type] = {}
            detour_time = getattr(cb, "_hook_time", DetourTime.NONE)
            if detour_time not in self.custom_callbacks[cb_type]:
                self.custom_callbacks[cb_type][detour_time] = {
                    cb,
                }
            else:
                self.custom_callbacks[cb_type][detour_time].add(cb)

    def _remove_custom_callbacks(self, callbacks: set[HookProtocol]):
        # Remove the values in the list which correspond to the data in `callbacks`
        for cb in callbacks:
            if (cb_type := cb._custom_trigger) is None:
                continue
            if cb_type in self.custom_callbacks:
                # Remove the functions from the set and then check whether it's
                # empty.
                self.custom_callbacks[cb_type][getattr(cb, "_hook_time", DetourTime.NONE)].discard(cb)
                if all(not x for x in self.custom_callbacks[cb_type].values()):
                    del self.custom_callbacks[cb_type]

    def call_custom_callbacks(self, callback_key: str, detour_time: DetourTime = DetourTime.NONE):
        """Call the specified custom callback with the given detour_time.

        Parameters
        ----------
        callback_key:
            The key which is used to reference the custom callback.
        detour_time:
            Whether to call the ``before`` or ``after`` detour.

        Notes
        -----
            If there is no callback registered for the key and detour_time combination nothing will happen.
        """
        callbacks = self.custom_callbacks.get(callback_key, {})
        if callbacks:
            for cb in callbacks.get(detour_time, set()):
                cb()

    def register_hook(self, hook: HookProtocol):
        """Register the provided hook."""
        hook_func_name = hook._hook_func_name
        # If the hook has an overload, add it here so that we can disambiguate them.
        if getattr(hook, "_func_overload", None) is not None:
            hook_func_name += f"({hook._func_overload})"
        # Check to see if this function hook name exists within the hook mapping.
        # If it doesn't then we need to initialise the FuncHook and then add the detour to it.
        # If it does then we simply add the detour to it.
        if hook_func_name not in self.hooks:
            # Check to see if it's a manual hook. If so, we initialize the FuncHook object differently.
            if hook._is_manual_hook:
                hook = cast(ManualHookProtocol, hook)
                funcdef = hook._hook_func_def
                if funcdef is None:
                    # If no funcdef is explicitly provided, look and see if we have one defined in the module
                    # data for this function name.
                    funcdef = module_data.FUNC_CALL_SIGS.get(hook_func_name)
                if funcdef is None:
                    raise SyntaxError(
                        "When creating a manual hook, the first detour for any given name MUST have a "
                        "func_def argument."
                    )
                # Handle hook offsets and patterns.
                hook_offset = hook._hook_offset
                if hook_offset is None:
                    hook_offset = module_data.FUNC_OFFSETS.get(hook_func_name)
                hook_pattern = hook._hook_pattern
                if hook_pattern is None:
                    hook_pattern = module_data.FUNC_PATTERNS.get(hook_func_name)
                if hook_offset is None and hook_pattern is None:
                    hook_logger.error(
                        f"The manual hook for {hook_func_name} was defined with no offset or pattern. One of "
                        "these is required to register a hook. The hook will not be registered."
                    )
                    return
                binary = hook._hook_binary or module_data.FUNC_BINARY
                self.hooks[hook_func_name] = FuncHook(
                    hook._hook_func_name,
                    offset=hook_offset,
                    pattern=hook_pattern,
                    func_def=funcdef,
                    binary=binary,
                )
            elif hook._is_imported_func_hook:
                hook = cast(HookProtocol, hook)
                dll_name = hook._dll_name.lower()
                hook_func_name = hook._hook_func_name
                hook_func_def = hook._hook_func_def
                if (dll_func_ptrs := _internal.imports.get(dll_name)) is not None:
                    func_ptr = dll_func_ptrs.get(hook_func_name)
                    # For now, cast the func_ptr object back to the target location in memory.
                    # This is wasteful, but simple for now for testing...
                    target = ctypes.cast(func_ptr, ctypes.c_void_p).value
                    self.hooks[hook_func_name] = FuncHook(
                        hook._hook_func_name,
                        offset=target,
                        func_def=hook_func_def,
                        offset_is_absolute=True,
                    )
                else:
                    hook_logger.error(f"Cannot find {dll_name} in the import list")
            elif hook._is_exported_func_hook:
                if _internal.BINARY_PATH is None:
                    hook_logger.error("Current running binary path unknown. Cannot hook exported functions")
                    return
                # TODO: This is inefficient. We should only instantiate the "dll" once.
                own_dll = ctypes.WinDLL(_internal.BINARY_PATH)
                func_ptr = getattr(own_dll, hook._hook_func_name)
                target = ctypes.cast(func_ptr, ctypes.c_void_p).value
                self.hooks[hook_func_name] = FuncHook(
                    hook._hook_func_name,
                    offset=target,
                    func_def=hook._hook_func_def,
                    offset_is_absolute=True,
                )
            else:
                # TODO: have a way to differentiate the binary here.
                self.hooks[hook_func_name] = FuncHook(
                    hook._hook_func_name,
                    offset=hook._hook_offset,
                    pattern=hook._hook_pattern,
                    func_def=hook._hook_func_def,
                    overload=getattr(hook, "_func_overload", None),
                )
            self._uninitialized_hooks.add(hook_func_name)
        self.hooks[hook_func_name].add_detour(hook)

        if getattr(hook, "_get_caller", False):
            self._get_caller_detours.add(hook_func_name)

    def initialize_hooks(self) -> int:
        """Initialize any uninitialized hooks.
        This will also enable the hooks so that they become active.
        """
        count = 0
        for hook_name in self._uninitialized_hooks:
            hook = self.hooks[hook_name]
            hook._init()
            bound = hook.bind()
            if bound:
                count += 1
            else:
                # If the mod didn't get bound, we don't try and enable it!
                continue
            # Try and enable the hook.
            try:
                hook.enable()
                if hook._offset_is_absolute:
                    offset = hook.target
                    prefix = ""
                else:
                    offset = hook.offset
                    prefix = f"{hook._binary}+"
                hook_logger.info(f"Enabled hook for {hook_name} at {prefix}0x{offset:X}")
            except Exception:
                hook_logger.error(f"Unable to enable {hook_name} because:")
                hook_logger.error(traceback.format_exc())

            # If any of the hooked functions want to log where they were called from, we need to overwrite
            # part of the trampoline bytes to capture the RSP register.
            if hook_name in self._get_caller_detours:
                if not HAS_ICED:
                    hook_logger.error(
                        f"Cannot get calling address of {hook_name} as `iced_x86` package is not installed.\n"
                        "Please install and try again."
                    )
                    continue
                # First, get the first jump so we can go to the trampoline bytes.
                jmp_data = (ctypes.c_char * 0x10).from_address(hook.target)
                jmp_addr = get_first_jmp_addr(jmp_data.raw, hook.target)
                if jmp_addr:
                    # minhook seems to have only gotten 0x20 bytes for the trampoline. Would be nice to have
                    # more but this is all we have to work with. It's luckily *just* enough.
                    # If we ever need more we'll need to make our own little detour somewhere else.
                    data_at_detour = (ctypes.c_char * 0x20).from_address(jmp_addr)

                    rsp_buff_addr = get_addressof(hook._rsp_addr)

                    rsp_load_bytes = generate_load_stack_pointer_bytes(rsp_buff_addr, jmp_addr, BITS)
                    # Get the original bytes written by minhook so that we can restore them.
                    orig_bytes = data_at_detour.raw[:0xE]
                    for i in range(len(rsp_load_bytes)):
                        data_at_detour[i] = rsp_load_bytes[i]
                    for j in range(len(orig_bytes)):
                        data_at_detour[i + j + 1] = orig_bytes[j]
                    hook_logger.info(
                        f"The function {hook_name} has a modified hook to get the calling address."
                    )

        # There are no uninitialized hooks.
        self._uninitialized_hooks = set()
        return count

    def enable(self, func_name: str):
        """Enable the hook for the provided function name."""
        if hook := self.hooks.get(func_name):
            hook.enable()
            hook_logger.info(f"Enabled hook for function '{func_name}'")
        else:
            hook_logger.error(f"Couldn't enable hook for function '{func_name}'")
            return

    def disable(self, func_name: str):
        """Disable the hook for the provided function name."""
        if hook := self.hooks.get(func_name):
            hook.disable()
            hook_logger.info(f"Disabled hook for function '{func_name}'")
        else:
            hook_logger.error(f"Couldn't disable hook for function '{func_name}'")
            return

    def _debug_show_states(self):
        # Return the states of all the registered hooks
        for func_name, hook in self.hooks.items():
            hook_logger.info(f"Functions registered for {func_name}:")
            if hook._before_detours:
                hook_logger.info("  Before Detours:")
                for func in hook._before_detours:
                    hook_logger.info(f"    {func}")
            if hook._after_detours:
                hook_logger.info("  After Detours:")
                for func in hook._after_detours:
                    hook_logger.info(f"    {func}")


class Structure(ctypes.Structure):
    """Simple wrapper around ctypes.Structure."""

    def __getattribute__(self, name: str):
        # Hook the instance attribute lookup so that we may "bind" the instance to the returned FunctionHook
        # instance.
        # We need to do this because the decorator has no knowledge of the actual bound instance at run-time.
        res = super().__getattribute__(name)
        if isinstance(res, FunctionHook):
            res._bound_class = self
        return res


P = ParamSpec("P")
R = TypeVar("R")
S = TypeVar("S", bound=Structure)
THIS = TypeVar(
    "THIS",
    bound=Union[
        ctypes.c_uint32,  # 32 bit "pointer" types.
        ctypes.c_ulong,
        ctypes.c_uint64,  # 64 bit "pointer" types.
        ctypes.c_ulonglong,
        ctypes._Pointer,  # Actual pointer type.
    ],
)


class FunctionHook(Generic[P, R]):
    def __init__(
        self,
        func: Union[Callable[P, R], Callable[Concatenate[S, THIS, P], R]],
        signature: Optional[str] = None,
        offset: Optional[int] = None,
        exported_name: Optional[str] = None,
        imported_name: Optional[str] = None,
        overload_id: Optional[str] = None,
        is_static: bool = False,
    ):
        self._func = func
        self._signature = signature
        self._offset = offset
        self._exported_name = exported_name
        self._imported_name = imported_name
        self._overload_id = overload_id
        self._is_static = is_static
        self._this_is_pointer: Optional[bool] = None
        self._bound_class: Optional[ctypes.Structure] = None
        self._funcdef: Optional[FuncDef] = None

    @property
    def this_is_pointer(self):
        """Only valid for bound methods. Returns True if the first argument is a pointer type."""
        if self._this_is_pointer is not None:
            return self._this_is_pointer
        if self._funcdef is None:
            self._funcdef = _get_funcdef(self._func)
        self._this_is_pointer = issubclass(self._funcdef.arg_types[0], ctypes._Pointer)
        return self._this_is_pointer

    def _call(self, *args, **kwargs) -> Optional[R]:
        """Call the actual function. This will do some work to find where the function is in memory and then
        call it with the provided arguments.
        """
        try:
            # Get the FUNCDEF. This will have named arguments with types so that we may construct a function
            # prototype which allows kwargs.
            if self._funcdef is None:
                self._funcdef = _get_funcdef(self._func)
            # Unfortunately the ctypes function prototype can only be called with kwargs if its a function
            # which is in a remote library.
            # We can do this for imported and exported functions, but to have the logic the same for all, it's
            # better to just flatten the kwargs and args into a single set of args and pass into the function
            # prototype defined by an offset.
            _args = self._funcdef.flatten(*args, **kwargs)
            sig = CFUNCTYPE(self._funcdef.restype, *self._funcdef.arg_types)
            binary_base = _internal.BASE_ADDRESS
            # Depending on what kind of function we are calling, we change how we find the offset of the func.
            offset = None
            if self._offset is not None:
                offset = binary_base + self._offset
            elif self._signature is not None:
                rel_offset = find_pattern_in_binary(self._signature, False, _internal.EXE_NAME)
                if rel_offset is not None and isinstance(rel_offset, int):
                    offset = binary_base + rel_offset
            elif self._exported_name is not None:
                own_dll = ctypes.WinDLL(_internal.BINARY_PATH)
                func_ptr = getattr(own_dll, self._exported_name)
                offset = ctypes.cast(func_ptr, ctypes.c_void_p).value
            # Finally, call the function.
            if offset is not None:
                cfunc = sig(offset)
                val = cfunc(*_args)
                return val
            else:
                hook_logger.error(f"Unable to call {self._func.__qualname__!r} - Cannot find function.")
        except Exception:
            hook_logger.exception(f"There was an exception calling {self._func.__qualname__!r}")
            return None

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> Optional[R]:
        # This initial check is to check if the first argument was a function.
        # This will only happen if the function is being used as a decorator.
        # if this check fails, then we are calling the function under "normal" usage.
        if args and inspect.isfunction(args[0]):
            # In this case the decorator was used without a .before or .after -> raise error
            raise ValueError(
                f"Hook for detour {self._func.__qualname__!r} must be specified as either `before` or `after`"
            )

        if self._is_static:
            # For a static method, we don't need to worry about any binding, we can just call it with the
            # provided arguments.
            return self._call(*args, **kwargs)
        else:
            # For a non-static method, we need to do more work since we need to get the instance the method is
            # bound to, and then get the address of it and pass it in as the first argument.
            if self._bound_class is not None:
                try:
                    if self._this_is_pointer:
                        return self._call(ctypes.byref(self._bound_class), *args, **kwargs)
                    else:
                        # If it's not a pointer, then we'll assume it's an int and pass the address...
                        return self._call(ctypes.addressof(self._bound_class), *args, **kwargs)
                except Exception:
                    hook_logger.exception(f"Failed to call {self._func.__qualname__} with args {args}")
            else:
                raise ValueError("Not bound to anything...")

    def _decorate_detour(self, detour, hook_time: Optional[DetourTime] = None) -> HookProtocol:
        if hook_time is None:
            raise ValueError(
                f"Hook for detour {detour.__qualname__!r} must be specified as either `before` or `after`"
            )

        self._funcdef = _get_funcdef(self._func)

        setattr(detour, "_is_funchook", True)
        setattr(detour, "_hook_time", hook_time)
        if self._exported_name is None:
            setattr(detour, "_hook_func_name", self._func.__qualname__)
        else:
            setattr(detour, "_hook_func_name", self._exported_name)
        if self._imported_name is not None:
            split_name = self._imported_name.split(".", maxsplit=1)
            if len(split_name) != 2:
                raise ValueError(
                    f"imported name {self._imported_name!r} is invalid. Please ensure it has the following "
                    "structure: dll_name.dll_function"
                )
            dll_name, function_name = split_name
            setattr(detour, "_dll_name", dll_name)
            setattr(detour, "_is_imported_func_hook", True)
            setattr(detour, "_hook_func_name", function_name)
        else:
            setattr(detour, "_is_imported_func_hook", False)
            setattr(detour, "_dll_name", None)
        setattr(detour, "_hook_func_def", self._funcdef.to_FUNCDEF())
        setattr(detour, "_hook_offset", self._offset)
        setattr(detour, "_hook_pattern", self._signature)
        setattr(detour, "_is_manual_hook", False)
        setattr(detour, "_is_exported_func_hook", self._exported_name is not None)
        setattr(detour, "_has__result_", False)
        setattr(detour, "_noop", False)
        setattr(detour, "_func_overload", self._overload_id)
        return detour

    def after(self, detour: Callable) -> HookProtocol:
        """Mark the detour as running after the original function."""
        decorated_detour = self._decorate_detour(detour, DetourTime.AFTER)
        if "_result_" in inspect.signature(detour).parameters.keys():
            setattr(detour, "_has__result_", True)
        return decorated_detour

    def before(self, detour: Callable) -> HookProtocol:
        """Mark the detour as running before the original function."""
        decorated_detour = self._decorate_detour(detour, DetourTime.BEFORE)
        return decorated_detour

    def overload(self, overload_id: str) -> Self:
        """Get an instance of the class which corresponds to the specified overload id.
        This overload id should be provided as the ``overload_id`` argument for ``function_hook``"""
        if overload_id == self._overload_id:
            return self
        else:
            fh = _FunctionHook_overloads.get(self._func.__qualname__, {}).get(overload_id, None)
            if fh is not None:
                return fh
            else:
                raise ValueError(f"Unknown overload {overload_id!r} for {self._func.__qualname__}")


class _function_hook:
    def __init__(
        self,
        signature: Optional[str] = None,
        offset: Optional[int] = None,
        exported_name: Optional[str] = None,
        imported_name: Optional[str] = None,
        overload_id: Optional[str] = None,
    ):
        self.signature = signature
        self.offset = offset
        self.exported_name = exported_name
        self.imported_name = imported_name
        self.overload_id = overload_id


class static_function_hook(_function_hook):
    """Mark the decorated function as a static function hook.

    .. note::
        Only of the arguments of this function is required.
        The order the arguments are respected is ``signature``, ``offset``, then ``exported_name``.

    .. note::
        You do not need to apply the ``@staticmethod`` decorator to functions if you use this decorator,
        however your static type checker may complain, so this decorator is safe to apply on top of the
        ``@staticmethod`` decorator.

    Parameters
    ----------
    signature:
        A string representing the bytes which can be used to uniquely find the function within the binary.
    offset:
        The relative offset within the binary where the start of the function can be found.
    exported_name:
        The name of the exported function which is to be hooked.

        .. note::
            It is recommended that the function name is the "mangled" version.
            Ie. do not "demangle" the function name.

    imported_name:
        The full name of the function within the imported dll to be hooked. For example, this could be
        ``"Kernel32.ReadFile"``.
    overload_id:
        A unique name within each set of overloaded functions which can be used to identify the overload for
        calling and hooking purposes.
    """

    def __call__(self, func: Callable[P, R]) -> FunctionHook[P, R]:
        if not self.signature and not self.offset and not self.exported_name and not self.imported_name:
            raise ValueError(
                f"One of the `function_hook` arguments must be provided for {func.__qualname__!r}"
            )
        # Pass the function in directly so that we may defer the usage of inspect until the actual decorator
        # is called.
        # This will mean that only functions which are used are inspected which will massively reduce the
        # amount of work required.
        if isinstance(func, staticmethod):
            # Unwrap the static method to get the underlying function since staticmethods aren't callable
            # cf. https://bugs.python.org/issue20309
            func = func.__func__
        return FunctionHook[P, R](
            func,
            self.signature,
            self.offset,
            self.exported_name,
            self.imported_name,
            is_static=True,
        )


class function_hook(_function_hook):
    """Mark the decorated function as a function hook.

    .. note::
        Only of the arguments of this function is required.
        The order the arguments are respected is ``signature``, ``offset``, then ``exported_name``.

    .. important::
        This decorator must only be applied to non-static methods.
        This means that the first two arguments MUST be `self` (the usual python one), and `this` (the c
        one.)
        Because of how this decorator works, the function arguments will be determined from all the
        arguments proceeding these two mandatory ones.

    .. important::
        For this decorator to work, the class the method belongs to MUST be a
        :class:`~pymhf.core.hooking.Structure` instead of the usual `ctypes.Structure`.

    Parameters
    ----------
    signature:
        A string representing the bytes which can be used to uniquely find the function within the binary.
    offset:
        The relative offset within the binary where the start of the function can be found.
    exported_name:
        The name of the exported function which is to be hooked.

        .. note::
            It is recommended that the function name is the "mangled" version.
            Ie. do not "demangle" the function name.

    imported_name:
        The full name of the function within the imported dll to be hooked. For example, this could be
        ``"Kernel32.ReadFile"``.
    overload_id:
        A unique name within each set of overloaded functions which can be used to identify the overload for
        calling and hooking purposes.
    """

    def __call__(self, func: Callable[Concatenate[S, THIS, P], R]) -> FunctionHook[P, R]:
        if not self.signature and not self.offset and not self.exported_name and not self.imported_name:
            raise ValueError(
                f"One of the `function_hook` arguments must be provided for {func.__qualname__!r}"
            )
        # Pass the function in directly so that we may defer the usage of inspect until the actual decorator
        # is called.
        # This will mean that only functions which are used are inspected which will massively reduce the
        # amount of work required.
        fh = FunctionHook[P, R](
            func,
            self.signature,
            self.offset,
            self.exported_name,
            self.imported_name,
            self.overload_id,
            is_static=False,
        )
        _FunctionHook_overloads[func.__qualname__][self.overload_id] = fh
        return fh


hook_manager = HookManager()



# A collection of functions which will hash and cache things.

import hashlib
import json
import os
import os.path as op
from io import BufferedReader
from typing import Optional

from pymem.ressources.structure import MODULEINFO

import pymhf.core._internal as _internal

CACHE_DIR = ".cache"


def hash_bytes(fileobj: BufferedReader, _bufsize: int = 2**18) -> str:
    # Essentially implement hashlib.file_digest since it's python 3.11+
    # cf. https://github.com/python/cpython/blob/main/Lib/hashlib.py#L195
    digestobj = hashlib.sha1()
    buf = bytearray(_bufsize)  # Reusable buffer to reduce allocations.
    view = memoryview(buf)
    while True:
        size = fileobj.readinto(buf)
        if size == 0:
            break  # EOF
        digestobj.update(view[:size])
    return digestobj.hexdigest()


class OffsetCache:
    def __init__(self, path: str):
        self._path = path
        self._binary_hash: Optional[str] = None
        self._lookup: dict[str, int] = {}

    @property
    def path(self) -> str:
        return op.join(_internal.CWD, CACHE_DIR, f"{self._binary_hash}_{self._path}.json")

    def load(self, binary_hash: str):
        """Load the data."""
        self._binary_hash = binary_hash
        if op.exists(self.path):
            with open(self.path, "r") as f:
                self._lookup = json.load(f)

    def save(self):
        """Persist the cache to disk."""
        with open(self.path, "w") as f:
            json.dump(self._lookup, f)

    def get(self, name: str) -> Optional[int]:
        """Get the offset based on the key provided."""
        return self._lookup.get(name)

    def set(self, key: str, value: int, save: bool = True):
        """Set the key with the given value and optionally save."""
        self._lookup[key] = value
        if save:
            self.save()

    def items(self):
        for key, value in self._lookup.items():
            yield key, value


function_cache = OffsetCache("function_cache")
pattern_cache = OffsetCache("pattern_cache")
globals_cache = OffsetCache("globals_cache")
builtins_cache = OffsetCache("builtins_cache")


def load_caches(binary_hash: str):
    if not op.exists(op.join(_internal.CWD, CACHE_DIR)):
        os.makedirs(op.join(_internal.CWD, CACHE_DIR))
    function_cache.load(binary_hash)
    pattern_cache.load(binary_hash)
    globals_cache.load(binary_hash)
    builtins_cache.load(binary_hash)


module_map: dict[str, MODULEINFO] = {}



# Some extenions to ctypes to add some extra functionality

import ctypes
import types
from enum import IntEnum
from typing import Generic, Type, TypeVar

_cenum_type_cache = {}

T = TypeVar("T", bound=IntEnum)


class c_enum32(ctypes.c_int32, Generic[T]):
    """c_int32 wrapper for enums. This doesn't have the full set of features an enum would normally have,
    but just enough to make it useful."""

    _enum_type: Type[T]

    @property
    def _enum_value(self):
        return self._enum_type(self.value)

    @property
    def name(self) -> str:
        return self._enum_value.name

    def __str__(self):
        return self._enum_value.__str__()

    def __repr__(self):
        return self._enum_value.__repr__()

    def __eq__(self, other):
        return other == self.value

    def __class_getitem__(cls: Type["c_enum32"], enum_type: Type[T]):
        """Get the actual concrete type based on the enum_type provided.
        This will be cached so we only generate one instance of the type per IntEnum."""
        if not issubclass(enum_type, IntEnum):
            raise TypeError(f"Indexed type {enum_type!r} of type {enum_type} is not an IntEnum")
        if enum_type in _cenum_type_cache:
            return _cenum_type_cache[enum_type]
        else:
            _cls: Type[c_enum32[T]] = types.new_class(f"c_enum32[{enum_type}]", (c_enum32,))
            _cls._enum_type = enum_type
            _cenum_type_cache[enum_type] = _cls
            return _cls



import _winapi
import ctypes
import ctypes.wintypes

kernel32 = ctypes.WinDLL("kernel32.dll")

# Map two useful debugging functions in kernel32:

# Start debugger (will stop the current process)
# https://learn.microsoft.com/en-us/windows/win32/api/debugapi/nf-debugapi-debugactiveprocess
DebugActiveProcess = kernel32.DebugActiveProcess
DebugActiveProcess.argtypes = (ctypes.c_long,)
DebugActiveProcess.restype = ctypes.c_byte

# Stop the debugger (will result the current process)
# https://learn.microsoft.com/en-us/windows/win32/api/debugapi/nf-debugapi-debugactiveprocessstop
DebugActiveProcessStop = kernel32.DebugActiveProcessStop
DebugActiveProcessStop.argtypes = (ctypes.c_long,)
DebugActiveProcessStop.restype = ctypes.c_byte


class SECURITY_ATTRIBUTES(ctypes.Structure):
    _fields_ = [
        ("nLength", ctypes.wintypes.DWORD),
        ("lpSecurityDescriptor", ctypes.wintypes.LPVOID),
        ("bInheritHandle", ctypes.wintypes.BOOL),
    ]


def start_process(binary_path: str, creationflags: int = 0x0):
    # Start an executable similarly to subprocess.Popen
    # The functionality here is ripped directly from that implementation,
    # however we don't discard the thread_handle here so that we may use it
    # again later to resume the thread if we pause it.
    proc_attr = SECURITY_ATTRIBUTES()
    proc_attr.bInheritHandle = True
    thread_attr = SECURITY_ATTRIBUTES()
    thread_attr.bInheritHandle = True
    # Start the process the internal way to get the thread handle.
    handle_process, handle_thread, pid, tid = _winapi.CreateProcess(
        None,
        binary_path,
        ctypes.byref(proc_attr),
        ctypes.byref(thread_attr),
        False,
        creationflags,
        None,
        None,
        None,
    )
    return (handle_process, handle_thread, pid, tid)


def _stop_process(pid: int):
    ret = DebugActiveProcess(pid)
    if not ret:
        return f"Error: {ctypes.GetLastError()}"
    else:
        return ret


def _start_process(pid: int):
    ret = DebugActiveProcessStop(pid)
    if not ret:
        return f"Error: {ctypes.GetLastError()}"
    else:
        return ret



import ctypes
import types
from typing import TYPE_CHECKING, Any, Generator, Generic, Type, TypeVar, Union

if TYPE_CHECKING:
    from ctypes import _Pointer

CTYPES = Union[ctypes._SimpleCData, ctypes.Structure, ctypes._Pointer]

T = TypeVar("T", bound=CTYPES)
N = TypeVar("N", bound=int)
T1 = TypeVar("T1", bound=CTYPES)
T2 = TypeVar("T2", bound=CTYPES)

# Good source of info for these:
# # https://devblogs.microsoft.com/search?query=inside+stl&blog=%2Foldnewthing%2F


class _array(ctypes.Structure, Generic[T, N]):
    _elements: list[T]

    def __class_getitem__(cls: Type["_array"], key: tuple[Type[T], int]):
        _type, _size = key
        _cls: Type[_array[T, N]] = types.new_class(f"std::aray<{_type}, {_size}>", (cls,))
        _cls._fields_ = [  # type: ignore
            ("_elements", _type * _size),
        ]
        return _cls

    def __len__(self) -> int:
        return len(self._elements)

    def __getitem__(self, i: int) -> T:
        return self._elements[i]

    def __setitem__(self, i: int, val: T):
        self._elements[i] = val

    def __iter__(self) -> Generator[T, None, None]:
        for i in range(len(self)):
            yield self[i]


class _vb_val(ctypes.Structure, Generic[T]):
    #   std::vector<unsigned int,StackAllocator<unsigned int,1,-1> > _Myvec;
    # unsigned __int64 _Mysize;
    pass


class _vector(ctypes.Structure, Generic[T]):
    _template_type: T
    if TYPE_CHECKING:
        _first: _Pointer[Any]
        _last: _Pointer[Any]
        _end: _Pointer[Any]

    def __class_getitem__(cls: Type["_vector"], type_: Type[T]):
        _cls: Type[_vector[T]] = types.new_class(f"std::vector<{type_}>", (cls,))
        _cls._template_type = type_
        _cls._fields_ = [  # type: ignore
            ("_first", ctypes.POINTER(type_)),
            ("_last", ctypes.POINTER(type_)),
            ("_end", ctypes.POINTER(type_)),
        ]
        return _cls

    def __len__(self) -> int:
        return (
            ctypes.addressof(self._last.contents) - ctypes.addressof(self._first.contents)
        ) // 0x8  # Assuming 64 bit architecture

    def __getitem__(self, i: int) -> T:
        return self._first[i]

    def __iter__(self) -> Generator[T, None, None]:
        for i in range(len(self)):
            yield self[i]

    def clear(self):
        """
        Empty the vector. This will remove all references to the elements.
        USE THIS WISELY as is may cause issues.
        """
        nullptr = ctypes.POINTER(self._template_type)
        self._first = nullptr()
        self._last = nullptr()
        self._end = nullptr()


class _pair(ctypes.Structure, Generic[T1, T2]):
    first: T1
    second: T2

    def __class_getitem__(cls: Type["_pair"], key: tuple[Type[T1], Type[T2]]):
        first, second = key
        _cls: Type[_pair[T1, T2]] = types.new_class(f"std::pair<{first}, {second}>", (cls,))
        _cls._fields_ = [  # type: ignore
            ("first", first),
            ("second", second),
        ]
        return _cls


# 0x40 long
class _unordered_map(ctypes.Structure, Generic[T1, T2]):
    """
    std::_Hash<
        std::_Umap_traits<
            TkID<128>,
            TkID<256>,
            std::_Uhash_compare<TkID<128>,
            TkIDUnorderedMap::Hash128,
            std::equal_to<TkID<128> >
        >,
        TkSTLAllocatorShim<
            std::pair<TkID<128> const ,TkID<256> >
            ,8,
            -1>
        ,
        0>
    >
    +0x00 std::_Umap_traits<TkID<128>,TkID<256>,std::_Uhash_compare<TkID<128>,TkIDUnorderedMap::Hash128,std::equal_to<TkID<128> > >,TkSTLAllocatorShim<std::pair<TkID<128> const ,TkID<256> >,8,-1>,0> _Traitsobj;
    +0x08 std::list<
            std::pair<
                TkID<128> const ,
                TkID<256>
            >,
            TkSTLAllocatorShim<
                std::pair<
                    TkID<128> const ,
                    TkID<256>
                >,
                8,
                -1
            >
          > _List;
    +0x18 std::_Hash_vec<TkSTLAllocatorShim<std::_List_unchecked_iterator<std::_List_val<std::_List_simple_types<std::pair<TkID<128> const ,TkID<256> > > > >,8,-1> > _Vec;
    +0x30 unsigned __int64 _Mask;
    +0x38 unsigned __int64 _Maxidx;
    """

    pass


class _list_node(ctypes.Structure, Generic[T]):
    """
    +0x00 std::_List_node<std::pair<TkID<128> const ,TkID<256> >,void *> *_Next;
    +0x08 std::_List_node<std::pair<TkID<128> const ,TkID<256> >,void *> *_Prev;
    +0x10 std::pair<TkID<128> const ,TkID<128> > _Myval;
    """

    if TYPE_CHECKING:
        _next: _Pointer["_list_node"]
        _myval: _pair


_list_node._fields_ = [
    ("_next", ctypes.POINTER(_list_node)),  # TEMP
    ("_prev", ctypes.POINTER(_list_node)),  # TEMP
    ("_myval", _pair),
]


# 0x10 long
class _list(ctypes.Structure, Generic[T]):
    """
    +0x00 std::_List_node<std::pair<TkID<128> const ,TkID<256> >,void *> *_Myhead;
    +0x08 unsigned __int64 _Mysize;
    """

    pass


class std:
    array = _array
    vector = _vector
    pair = _pair
    list = _list


# class STD:
#     class ARRAY():
#         def __new__(cls, type_: Type[T], size: int) -> std.array:
#             _cls = types.new_class(
#                 f"std::aray<{type_}, {size}>", (std.array,)
#             )
#             _cls._fields_ = [
#                 ("_elements", type_ * size),
#             ]
#             return _cls


if __name__ == "__main__":
    data = bytearray(b"\x01\x02\x00\x00\x07\x00\x00\x00")
    pear = std.pair[ctypes.c_uint32, ctypes.c_int32]
    k = pear.from_buffer(data)
    print(k.first)
    print(k.second)
    harry = std.array[ctypes.c_ubyte, 6]
    d = harry.from_buffer(data)
    for i in d:
        print(i)
    print("setting")
    d[3] = 9
    for i in d:
        print(i)

    # print("AAAAAAAAA")
    # arry = STD.ARRAY(ctypes.c_ubyte, 6)
    # pp = arry.from_buffer(data)
    # for i in pp:
    #     print(i)
